#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "api_metadata_container",
    "api_run",
    "api_model",
    "api_project",
    "api_model_version",
]

import random
import string
import uuid
from random import randint

from neptune.new.internal.backends.api_model import ApiExperiment
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.id_formats import SysId, UniqueId


def api_metadata_container(container_type: ContainerType) -> ApiExperiment:
    if container_type == ContainerType.PROJECT:
        return api_project()
    elif container_type == ContainerType.RUN:
        return api_run()
    elif container_type == ContainerType.MODEL:
        return api_model()
    elif container_type == ContainerType.MODEL_VERSION:
        return api_model_version()
    else:
        raise ValueError(f"Unknown container_type: {container_type.value}")


def api_project() -> ApiExperiment:
    return _api_metadata_container(
        sys_id=_random_key(), container_type=ContainerType.PROJECT
    )


def api_run() -> ApiExperiment:
    return _api_metadata_container(
        sys_id=f"{_random_key()}-{randint(42, 12342)}", container_type=ContainerType.RUN
    )


def api_model() -> ApiExperiment:
    return _api_metadata_container(
        sys_id=f"{_random_key()}-{_random_key()}", container_type=ContainerType.MODEL
    )


def api_model_version() -> ApiExperiment:
    return _api_metadata_container(
        sys_id=f"{_random_key()}-{_random_key()}-{randint(42, 12342)}",
        container_type=ContainerType.MODEL_VERSION,
    )


def _api_metadata_container(
    sys_id: str, container_type: ContainerType
) -> ApiExperiment:
    return ApiExperiment(
        id=UniqueId(str(uuid.uuid4())),
        type=container_type,
        sys_id=SysId(sys_id),
        workspace="workspace",
        project_name="sandbox",
        trashed=False,
    )


def _random_key(key_length: int = 3) -> str:
    return "".join(
        random.choice(string.ascii_letters).upper() for _ in range(key_length)
    )



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class MetricsContainer(object):
    def __init__(
        self, cpu_usage_metric, memory_metric, gpu_usage_metric, gpu_memory_metric
    ):
        self.cpu_usage_metric = cpu_usage_metric
        self.memory_metric = memory_metric
        self.gpu_usage_metric = gpu_usage_metric
        self.gpu_memory_metric = gpu_memory_metric

    def metrics(self):
        return [
            metric
            for metric in [
                self.cpu_usage_metric,
                self.memory_metric,
                self.gpu_usage_metric,
                self.gpu_memory_metric,
            ]
            if metric is not None
        ]



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access

from mock import MagicMock, patch

from neptune.new.attributes.series.string_series import StringSeries
from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


@patch("time.time", new=TestAttributeBase._now)
class TestStringSeries(TestAttributeBase):
    def test_assign_type_error(self):
        values = [55, "string", None]
        for value in values:
            with self.assertRaises(Exception):
                StringSeries(MagicMock(), MagicMock()).assign(value)

    def test_get(self):
        exp, path = self._create_run(), self._random_path()
        var = StringSeries(exp, path)
        var.log("asdfhadh")
        var.log("hej!")
        self.assertEqual("hej!", var.fetch_last())

    def test_log(self):
        exp, path = self._create_run(), self._random_path()
        var = StringSeries(exp, path)
        var.log([str(val) for val in range(0, 5000)])
        self.assertEqual("4999", var.fetch_last())
        values = list(var.fetch_values()["value"].array)
        expected = list(range(0, 5000))
        self.assertEqual(len(set(expected)), len(set(values)))



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

import neptune.new as neptune
from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from e2e_tests.utils import (
    initialize_container,
    reinitialize_container,
    with_check_if_file_appears,
)
from neptune.new.exceptions import NeptuneModelKeyAlreadyExistsError
from neptune.new.metadata_containers import Model
from neptune.new.project import Project


class TestInitRun(BaseE2ETest):
    def test_custom_run_id(self, environment):
        custom_run_id = "-".join((fake.word() for _ in range(3)))
        run = neptune.init_run(custom_run_id=custom_run_id, project=environment.project)

        key = self.gen_key()
        val = fake.word()
        run[key] = val
        run.sync()

        run.stop()

        exp2 = neptune.init_run(
            custom_run_id=custom_run_id, project=environment.project
        )
        assert exp2[key].fetch() == val

    def test_send_source_code(self, environment):
        exp = neptune.init_run(
            source_files="**/*.py",
            name="E2e init source code",
            project=environment.project,
        )

        # download sources
        exp.sync()
        with with_check_if_file_appears("files.zip"):
            exp["source_code/files"].download()


class TestInitProject(BaseE2ETest):
    def test_resuming_project(self, environment):
        exp = neptune.init_project(name=environment.project)

        key = self.gen_key()
        val = fake.word()
        exp[key] = val
        exp.sync()

        exp.stop()

        exp2 = neptune.init_project(name=environment.project)
        assert exp2[key].fetch() == val

    def test_init_and_readonly(self, environment):
        project: Project = neptune.init_project(name=environment.project)

        key = f"{self.gen_key()}-" + "-".join((fake.word() for _ in range(4)))
        val = fake.word()
        project[key] = val
        project.sync()
        project.stop()

        read_only_project = neptune.get_project(name=environment.project)
        read_only_project.sync()

        assert set(read_only_project.get_structure()["sys"]) == {
            "creation_time",
            "id",
            "modification_time",
            "monitoring_time",
            "name",
            "ping_time",
            "running_time",
            "size",
            "state",
            "tags",
            "visibility",
        }
        assert read_only_project[key].fetch() == val


class TestInitModel(BaseE2ETest):
    @pytest.mark.parametrize("container", ["model"], indirect=True)
    def test_fail_reused_model_key(self, container: Model, environment):
        with pytest.raises(NeptuneModelKeyAlreadyExistsError):
            model_key = container["sys/id"].fetch().split("-")[1]
            neptune.init_model(key=model_key, project=environment.project)


class TestReinitialization(BaseE2ETest):
    @pytest.mark.parametrize("container_type", AVAILABLE_CONTAINERS)
    def test_resuming_container(self, container_type, environment):
        container = initialize_container(
            container_type=container_type, project=environment.project
        )
        sys_id = container["sys/id"].fetch()

        key = self.gen_key()
        val = fake.word()
        container[key] = val
        container.sync()
        container.stop()

        reinitialized = reinitialize_container(
            sys_id=sys_id,
            container_type=container.container_type.value,
            project=environment.project,
        )
        assert reinitialized[key].fetch() == val



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from e2e_tests.base import BaseE2ETest, fake
from e2e_tests.utils import Environment, a_project_name
from neptune.management import (
    add_project_member,
    create_project,
    delete_project,
    get_project_list,
    get_project_member_list,
    get_workspace_member_list,
    remove_project_member,
)
from neptune.management.exceptions import (
    UserAlreadyHasAccess,
    UserNotExistsOrWithoutAccess,
)
from neptune.management.internal.utils import normalize_project_name


@pytest.mark.management
class TestManagement(BaseE2ETest):
    def test_standard_scenario(self, environment: Environment):
        project_name, project_key = a_project_name(project_slug=f"{fake.slug()}-mgmt")
        project_identifier = normalize_project_name(
            name=project_name, workspace=environment.workspace
        )

        assert project_identifier not in get_project_list(
            api_token=environment.admin_token
        )
        assert environment.user in get_workspace_member_list(
            name=environment.workspace, api_token=environment.admin_token
        )
        assert (
            get_workspace_member_list(
                name=environment.workspace, api_token=environment.admin_token
            ).get(environment.user)
            == "member"
        )

        created_project_identifier = create_project(
            name=project_name,
            key=project_key,
            visibility="priv",
            workspace=environment.workspace,
            api_token=environment.admin_token,
        )

        assert created_project_identifier == project_identifier
        assert created_project_identifier in get_project_list(
            api_token=environment.admin_token
        )
        assert environment.user not in get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

        add_project_member(
            name=created_project_identifier,
            username=environment.user,
            role="contributor",
            api_token=environment.admin_token,
        )

        assert environment.user in get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )
        assert (
            get_project_member_list(
                name=created_project_identifier, api_token=environment.admin_token
            ).get(environment.user)
            == "contributor"
        )
        assert created_project_identifier in get_project_list(
            api_token=environment.user_token
        )

        remove_project_member(
            name=created_project_identifier,
            username=environment.user,
            api_token=environment.admin_token,
        )

        assert environment.user not in get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

        delete_project(
            name=created_project_identifier, api_token=environment.admin_token
        )

        assert project_identifier not in get_project_list(
            api_token=environment.admin_token
        )

    def test_visibility_workspace(self, environment: "Environment"):
        project_name, project_key = a_project_name(
            project_slug=f"{fake.slug()}-workspace"
        )
        project_identifier = normalize_project_name(
            name=project_name, workspace=environment.workspace
        )

        assert project_identifier not in get_project_list(
            api_token=environment.admin_token
        )
        assert environment.user in get_workspace_member_list(
            name=environment.workspace, api_token=environment.admin_token
        )
        assert (
            get_workspace_member_list(
                name=environment.workspace, api_token=environment.admin_token
            ).get(environment.user)
            == "member"
        )

        created_project_identifier = create_project(
            name=project_name,
            key=project_key,
            visibility="workspace",
            workspace=environment.workspace,
            api_token=environment.admin_token,
        )

        assert created_project_identifier == project_identifier
        assert created_project_identifier in get_project_list(
            api_token=environment.admin_token
        )
        assert environment.user in get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )
        assert (
            get_project_member_list(
                name=created_project_identifier, api_token=environment.admin_token
            ).get(environment.user)
            == "owner"
        )

        with pytest.raises(UserAlreadyHasAccess):
            add_project_member(
                name=created_project_identifier,
                username=environment.user,
                role="contributor",
                api_token=environment.admin_token,
            )

        assert created_project_identifier in get_project_list(
            api_token=environment.user_token
        )

        with pytest.raises(UserNotExistsOrWithoutAccess):
            remove_project_member(
                name=created_project_identifier,
                username=environment.user,
                api_token=environment.admin_token,
            )

        delete_project(
            name=created_project_identifier, api_token=environment.admin_token
        )

        assert project_identifier not in get_project_list(
            api_token=environment.admin_token
        )



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
import random
import unittest
import uuid
from collections import namedtuple
from tempfile import NamedTemporaryFile, TemporaryDirectory

import mock
from mock import MagicMock, call, patch

from neptune.new.internal.backends.api_model import ClientConfig
from neptune.new.internal.backends.hosted_file_operations import (
    _get_content_disposition_filename,
    download_file_attribute,
    download_file_set_attribute,
    upload_file_attribute,
    upload_file_set_attribute,
)
from neptune.utils import IS_WINDOWS
from tests.neptune.new.backend_test_mixin import BackendTestMixin
from tests.neptune.new.utils.file_helpers import create_file


def set_expected_result(endpoint: MagicMock, value: dict):
    endpoint.return_value.response.return_value.result = namedtuple(
        endpoint.__class__.__name__, value.keys()
    )(**value)


class HostedFileOperationsHelper(unittest.TestCase):
    @staticmethod
    def get_random_bytes(count):
        return bytes(random.randint(0, 255) for _ in range(count))

    @staticmethod
    def _get_swagger_mock():
        swagger_mock = MagicMock()
        swagger_mock.swagger_spec.http_client = MagicMock()
        swagger_mock.swagger_spec.api_url = "ui.neptune.ai"
        swagger_mock.api.uploadFileSetAttributeChunk.operation.path_name = (
            "/uploadFileSetChunk"
        )
        swagger_mock.api.uploadFileSetAttributeTar.operation.path_name = (
            "/uploadFileSetTar"
        )
        swagger_mock.api.uploadPath.operation.path_name = "/uploadPath"
        swagger_mock.api.uploadAttribute.operation.path_name = "/attributes/upload"
        swagger_mock.api.downloadAttribute.operation.path_name = "/attributes/download"
        swagger_mock.api.downloadFileSetAttributeZip.operation.path_name = (
            "/attributes/downloadFileSetZip"
        )
        swagger_mock.api.downloadFileSetAttributeZip.operation.path_name = (
            "/attributes/downloadFileSetZip"
        )
        swagger_mock.api.download.operation.path_name = "/download"

        swagger_mock.api.fileAtomMultipartUploadStart.operation.path_name = (
            "/attributes/storage/file/upload/start"
        )
        swagger_mock.api.fileAtomMultipartUploadFinish.operation.path_name = (
            "/attributes/storage/file/upload/finish"
        )
        swagger_mock.api.fileAtomMultipartUploadPart.operation.path_name = (
            "/attributes/storage/file/upload/part"
        )
        swagger_mock.api.fileAtomUpload.operation.path_name = (
            "/attributes/storage/file/upload"
        )

        swagger_mock.api.fileSetFileMultipartUploadStart.operation.path_name = (
            "/attributes/storage/fileset/upload/start"
        )
        swagger_mock.api.fileSetFileMultipartUploadFinish.operation.path_name = (
            "/attributes/storage/fileset/upload/finish"
        )
        swagger_mock.api.fileSetFileMultipartUploadPart.operation.path_name = (
            "/attributes/storage/fileset/upload/part"
        )
        swagger_mock.api.fileSetFileUpload.operation.path_name = (
            "/attributes/storage/fileset/upload"
        )
        return swagger_mock


class TestCommonHostedFileOperations(HostedFileOperationsHelper):
    # pylint:disable=protected-access
    def test_get_content_disposition_filename(self):
        # given
        response_mock = MagicMock()
        response_mock.headers = {
            "Content-Disposition": 'attachment; filename="sample.file"'
        }

        # when
        filename = _get_content_disposition_filename(response_mock)

        # then
        self.assertEqual(filename, "sample.file")

    @patch(
        "neptune.new.internal.backends.hosted_file_operations._store_response_as_file"
    )
    @patch("neptune.new.internal.backends.hosted_file_operations._download_raw_data")
    def test_download_file_attribute(self, download_raw, store_response_mock):
        # given
        swagger_mock = self._get_swagger_mock()
        exp_uuid = str(uuid.uuid4())

        # when
        download_file_attribute(
            swagger_client=swagger_mock,
            container_id=exp_uuid,
            attribute="some/attribute",
            destination=None,
        )

        # then
        download_raw.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/attributes/download",
            headers={"Accept": "application/octet-stream"},
            query_params={"experimentId": str(exp_uuid), "attribute": "some/attribute"},
        )
        store_response_mock.assert_called_once_with(download_raw.return_value, None)

    @patch(
        "neptune.new.internal.backends.hosted_file_operations._store_response_as_file"
    )
    @patch("neptune.new.internal.backends.hosted_file_operations._download_raw_data")
    @patch(
        "neptune.new.internal.backends.hosted_file_operations._get_download_url",
        new=lambda _, _id: "some_url",
    )
    def test_download_file_set_attribute(self, download_raw, store_response_mock):
        # given
        swagger_mock = self._get_swagger_mock()
        download_id = str(uuid.uuid4())

        # when
        download_file_set_attribute(
            swagger_client=swagger_mock, download_id=download_id, destination=None
        )

        # then
        download_raw.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="some_url",
            headers={"Accept": "application/zip"},
        )
        store_response_mock.assert_called_once_with(download_raw.return_value, None)


class TestOldUploadFileOperations(HostedFileOperationsHelper):
    multipart_config = None

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    def test_missing_files_or_directory(self, upload_raw_data_mock):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_raw_data_mock.return_value = b"null"
        swagger_mock.api.getUploadConfig.return_value.response.return_value.result.chunkSize = (
            10
        )

        # when
        with NamedTemporaryFile("w") as temp_file_1:
            with NamedTemporaryFile("w") as temp_file_2:
                with TemporaryDirectory() as temp_dir:
                    upload_file_set_attribute(
                        swagger_client=swagger_mock,
                        container_id=exp_uuid,
                        attribute="some/attribute",
                        file_globs=[
                            temp_file_1.name,
                            temp_file_2.name,
                            os.path.abspath("missing_file"),
                            temp_dir,
                        ],
                        reset=True,
                        multipart_config=self.multipart_config,
                    )

        # then
        upload_raw_data_mock.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/uploadFileSetTar",
            data=mock.ANY,
            headers={"Content-Type": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True",
            },
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations._upload_loop")
    def test_upload_file_attribute(self, upload_loop_mock):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_loop_mock.return_value = b"null"

        # when
        with NamedTemporaryFile("w") as f:
            upload_file_attribute(
                swagger_client=swagger_mock,
                container_id=exp_uuid,
                attribute="target/path.txt",
                source=f.name,
                ext="txt",
                multipart_config=self.multipart_config,
            )

        # then
        upload_loop_mock.assert_called_once_with(
            file_chunk_stream=mock.ANY,
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/attributes/upload",
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "target/path.txt",
                "ext": "txt",
            },
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations._upload_loop")
    def test_upload_file_attribute_from_stream(self, upload_loop_mock):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_loop_mock.return_value = b"null"

        # when
        upload_file_attribute(
            swagger_client=swagger_mock,
            container_id=exp_uuid,
            attribute="target/path.txt",
            source=b"Some content of test stream",
            ext="txt",
            multipart_config=self.multipart_config,
        )

        # then
        upload_loop_mock.assert_called_once_with(
            file_chunk_stream=mock.ANY,
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/attributes/upload",
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "target/path.txt",
                "ext": "txt",
            },
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations._upload_loop_chunk")
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    def test_upload_single_file_in_file_set_attribute(self, upload_loop_chunk_mock):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_loop_chunk_mock.return_value = b"null"
        chunk_size = 5 * 1024 * 1024
        swagger_mock.api.getUploadConfig.return_value.response.return_value.result.chunkSize = (
            chunk_size
        )

        # when
        with NamedTemporaryFile("w") as temp_file:
            with open(temp_file.name, "wb") as handler:
                handler.write(self.get_random_bytes(2 * chunk_size))

            upload_file_set_attribute(
                swagger_client=swagger_mock,
                container_id=str(exp_uuid),
                attribute="some/attribute",
                file_globs=[temp_file.name],
                reset=True,
                multipart_config=self.multipart_config,
            )

        # then
        upload_loop_chunk_mock.assert_has_calls(
            [
                call(
                    mock.ANY,
                    mock.ANY,
                    http_client=swagger_mock.swagger_spec.http_client,
                    query_params={
                        "experimentId": str(exp_uuid),
                        "attribute": "some/attribute",
                        "reset": "True",
                        "path": os.path.basename(temp_file.name),
                    },
                    url="https://ui.neptune.ai/uploadFileSetChunk",
                ),
                call(
                    mock.ANY,
                    mock.ANY,
                    http_client=swagger_mock.swagger_spec.http_client,
                    query_params={
                        "experimentId": str(exp_uuid),
                        "attribute": "some/attribute",
                        "reset": "False",
                        "path": os.path.basename(temp_file.name),
                    },
                    url="https://ui.neptune.ai/uploadFileSetChunk",
                ),
            ]
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    def test_upload_multiple_files_in_file_set_attribute(self, upload_raw_data_mock):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_raw_data_mock.return_value = b"null"
        swagger_mock.api.getUploadConfig.return_value.response.return_value.result.chunkSize = (
            10
        )

        # when
        with NamedTemporaryFile("w") as temp_file_1:
            with NamedTemporaryFile("w") as temp_file_2:
                upload_file_set_attribute(
                    swagger_client=swagger_mock,
                    container_id=exp_uuid,
                    attribute="some/attribute",
                    file_globs=[temp_file_1.name, temp_file_2.name],
                    reset=True,
                    multipart_config=self.multipart_config,
                )

        # then
        upload_raw_data_mock.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/uploadFileSetTar",
            data=mock.ANY,
            headers={"Content-Type": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True",
            },
        )


class TestNewUploadFileOperations(HostedFileOperationsHelper, BackendTestMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config_swagger_client = self._get_swagger_client_mock(MagicMock())
        client_config = ClientConfig.from_api_response(
            config_swagger_client.api.getClientConfig().response().result
        )
        self.multipart_config = client_config.multipart_config

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    def test_missing_files_or_directory(self, upload_raw_data_mock):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_raw_data_mock.return_value = b"null"

        # when
        with NamedTemporaryFile("w") as temp_file_1:
            with NamedTemporaryFile("w") as temp_file_2:
                with TemporaryDirectory() as temp_dir:
                    upload_file_set_attribute(
                        swagger_client=swagger_mock,
                        container_id=exp_uuid,
                        attribute="some/attribute",
                        file_globs=[
                            temp_file_1.name,
                            temp_file_2.name,
                            os.path.abspath("missing_file"),
                            temp_dir,
                        ],
                        reset=True,
                        multipart_config=self.multipart_config,
                    )

        # then
        upload_raw_data_mock.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/uploadFileSetTar",
            data=mock.ANY,
            headers={"Content-Type": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True",
            },
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    def test_upload_small_file_attribute(self, upload_raw_data):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_raw_data.return_value = json.dumps(
            {
                "uploadId": "placeholder",
                "errors": [],
            }
        )
        data = b"testdata"

        # when
        with create_file(content=data, binary_mode=True) as filename:
            upload_file_attribute(
                swagger_client=swagger_mock,
                container_id=exp_uuid,
                attribute="target/path.txt",
                source=filename,
                ext="txt",
                multipart_config=self.multipart_config,
            )

        # then
        swagger_mock.api.fileSetFileMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileSetFileUpload.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileAtomUpload.assert_not_called()
        upload_raw_data.assert_called_once_with(
            data=data,
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/attributes/storage/file/upload",
            query_params={
                "experimentIdentifier": str(exp_uuid),
                "attribute": "target/path.txt",
                "ext": "txt",
            },
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    def test_upload_big_file_attribute(self, upload_raw_data):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_id = "placeholder"
        set_expected_result(
            swagger_mock.api.fileAtomMultipartUploadStart,
            {
                "uploadId": upload_id,
                "errors": [],
            },
        )
        upload_raw_data.return_value = json.dumps(
            {
                "errors": [],
            }
        )
        data = self.get_random_bytes(8 * 2**20)  # 8 MB
        chunk_size = self.multipart_config.min_chunk_size

        # when
        with create_file(content=data, binary_mode=True) as filename:
            upload_file_attribute(
                swagger_client=swagger_mock,
                container_id=exp_uuid,
                attribute="target/path.txt",
                source=filename,
                ext="txt",
                multipart_config=self.multipart_config,
            )

        # then
        swagger_mock.api.fileSetFileMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileSetFileUpload.assert_not_called()
        swagger_mock.api.fileAtomUpload.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadStart.assert_called_once_with(
            attribute="target/path.txt",
            experimentIdentifier=str(exp_uuid),
            ext="txt",
            totalLength=len(data),
        )
        swagger_mock.api.fileAtomMultipartUploadFinish.assert_called_once_with(
            attribute="target/path.txt",
            experimentIdentifier=str(exp_uuid),
            uploadId=upload_id,
        )
        upload_raw_data.assert_has_calls(
            [
                call(
                    data=data[:chunk_size],
                    http_client=swagger_mock.swagger_spec.http_client,
                    url="https://ui.neptune.ai/attributes/storage/file/upload/part",
                    headers={"X-Range": f"bytes=0-{chunk_size - 1}/{len(data)}"},
                    query_params={
                        "uploadPartIdx": 0,
                        "uploadId": upload_id,
                        "experimentIdentifier": str(exp_uuid),
                        "attribute": "target/path.txt",
                    },
                ),
                call(
                    data=data[chunk_size:],
                    http_client=swagger_mock.swagger_spec.http_client,
                    url="https://ui.neptune.ai/attributes/storage/file/upload/part",
                    headers={
                        "X-Range": f"bytes={chunk_size}-{len(data) - 1}/{len(data)}"
                    },
                    query_params={
                        "uploadPartIdx": 1,
                        "uploadId": upload_id,
                        "experimentIdentifier": str(exp_uuid),
                        "attribute": "target/path.txt",
                    },
                ),
            ]
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    def test_upload_single_small_file_in_file_set_attribute(self, upload_raw_data):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_raw_data.return_value = json.dumps(
            {
                "errors": [],
            }
        )
        data = b"testdata"

        # when
        with create_file(content=data, binary_mode=True) as filename:
            upload_file_set_attribute(
                swagger_client=swagger_mock,
                container_id=str(exp_uuid),
                attribute="some/attribute",
                file_globs=[filename],
                reset=True,
                multipart_config=self.multipart_config,
            )

        # then
        swagger_mock.api.fileSetFileMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileSetFileUpload.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileAtomUpload.assert_not_called()
        upload_raw_data.assert_called_once_with(
            data=data,
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/attributes/storage/fileset/upload",
            query_params={
                "subPath": os.path.basename(filename),
                "experimentIdentifier": str(exp_uuid),
                "attribute": "some/attribute",
            },
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    def test_upload_single_big_file_in_file_set_attribute(self, upload_raw_data):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_id = "placeholder"
        set_expected_result(
            swagger_mock.api.fileSetFileMultipartUploadStart,
            {
                "uploadId": upload_id,
                "errors": [],
            },
        )
        upload_raw_data.return_value = json.dumps(
            {
                "errors": [],
            }
        )
        data = self.get_random_bytes(8 * 2**20)  # 8 MB
        chunk_size = self.multipart_config.min_chunk_size

        # when
        with create_file(content=data, binary_mode=True) as filename:
            upload_file_set_attribute(
                swagger_client=swagger_mock,
                container_id=str(exp_uuid),
                attribute="some/attribute",
                file_globs=[filename],
                reset=True,
                multipart_config=self.multipart_config,
            )

        # then
        swagger_mock.api.fileSetFileMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileSetFileUpload.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileAtomUpload.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadStart.assert_called_once_with(
            attribute="some/attribute",
            experimentIdentifier=str(exp_uuid),
            totalLength=len(data),
            subPath=os.path.basename(filename),
        )
        swagger_mock.api.fileSetFileMultipartUploadFinish.assert_called_once_with(
            attribute="some/attribute",
            experimentIdentifier=str(exp_uuid),
            subPath=os.path.basename(filename),
            uploadId=upload_id,
        )
        upload_raw_data.assert_has_calls(
            [
                call(
                    data=data[:chunk_size],
                    http_client=swagger_mock.swagger_spec.http_client,
                    url="https://ui.neptune.ai/attributes/storage/fileset/upload/part",
                    headers={"X-Range": f"bytes=0-{chunk_size - 1}/{len(data)}"},
                    query_params={
                        "uploadPartIdx": 0,
                        "uploadId": upload_id,
                        "subPath": os.path.basename(filename),
                        "experimentIdentifier": str(exp_uuid),
                        "attribute": "some/attribute",
                    },
                ),
                call(
                    data=data[chunk_size:],
                    http_client=swagger_mock.swagger_spec.http_client,
                    url="https://ui.neptune.ai/attributes/storage/fileset/upload/part",
                    headers={
                        "X-Range": f"bytes={chunk_size}-{len(data) - 1}/{len(data)}"
                    },
                    query_params={
                        "uploadPartIdx": 1,
                        "uploadId": upload_id,
                        "subPath": os.path.basename(filename),
                        "experimentIdentifier": str(exp_uuid),
                        "attribute": "some/attribute",
                    },
                ),
            ]
        )

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch("neptune.new.internal.backends.hosted_file_operations.upload_raw_data")
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    def test_upload_multiple_files_in_file_set_attribute(self, upload_raw_data_mock):
        # given
        exp_uuid = str(uuid.uuid4())
        swagger_mock = self._get_swagger_mock()
        upload_raw_data_mock.return_value = b"null"
        swagger_mock.api.getUploadConfig.return_value.response.return_value.result.chunkSize = (
            10
        )

        # when
        with NamedTemporaryFile("w") as temp_file_1:
            with NamedTemporaryFile("w") as temp_file_2:
                upload_file_set_attribute(
                    swagger_client=swagger_mock,
                    container_id=exp_uuid,
                    attribute="some/attribute",
                    file_globs=[temp_file_1.name, temp_file_2.name],
                    reset=True,
                    multipart_config=self.multipart_config,
                )

        # then
        swagger_mock.api.fileSetFileMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileSetFileMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileSetFileUpload.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadStart.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadFinish.assert_not_called()
        swagger_mock.api.fileAtomMultipartUploadPart.assert_not_called()
        swagger_mock.api.fileAtomUpload.assert_not_called()
        upload_raw_data_mock.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="https://ui.neptune.ai/uploadFileSetTar",
            data=mock.ANY,
            headers={"Content-Type": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True",
            },
        )


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import unittest
from datetime import datetime

from neptune.new import (
    ANONYMOUS,
    Run,
    init,
    init_model,
    init_model_version,
    init_project,
    init_run,
)
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import (
    InactiveModelException,
    InactiveModelVersionException,
    InactiveProjectException,
    InactiveRunException,
    MetadataInconsistency,
    NeptuneProtectedPathException,
)
from neptune.new.metadata_containers import Model, ModelVersion, Project
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.series import FloatSeries, StringSeries


class TestExperiment(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    @classmethod
    def get_experiments(cls, flush_period=None):
        kwargs = {"mode": "debug"}
        if flush_period is not None:
            kwargs["flush_period"] = flush_period

        run1 = init(**kwargs)
        run2 = init_run(**kwargs)
        project = init_project(**kwargs)
        model = init_model(key="MOD", **kwargs)
        return [run1, run2, project, model]

    def test_define(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/value", Float(5), wait=True)
                self.assertEqual(
                    exp.get_structure()["some"]["path"]["value"].fetch(), 5
                )

    def test_define_string(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/value", String("Some string"), wait=True)
                self.assertEqual(
                    exp.get_structure()["some"]["path"]["value"].fetch(), "Some string"
                )

    def test_define_few_variables(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/num", Float(3))
                exp.define("some/path/text", String("Some text"), wait=True)
                self.assertEqual(exp.get_structure()["some"]["path"]["num"].fetch(), 3)
                self.assertEqual(
                    exp.get_structure()["some"]["path"]["text"].fetch(), "Some text"
                )

    def test_define_conflict(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/value", Float(5))
                with self.assertRaises(MetadataInconsistency):
                    exp.define("some/path/value", Float(1))

    def test_pop(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/num", Float(3))
                exp.define("some/path/text", String("Some text"))
                exp.pop("some/path/text")
                self.assertTrue("num" in exp.get_structure()["some"]["path"])
                self.assertTrue("text" not in exp.get_structure()["some"]["path"])

    def test_pop_namespace(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/subpath/num", Float(3))
                exp.define("some/path/text", String("Some text"))
                exp.define("some/otherpath", Float(4))
                exp.pop("some/path")
                self.assertTrue("path" not in exp.get_structure()["some"])

    def test_run_as_handler(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp.define("some/path/num", Float(3))
                exp.define("some/path/text", String("Some text"))
                handler = exp["some/path"]
                exp.wait()
                self.assertEqual(handler["num"].fetch(), 3)
                self.assertEqual(handler["text"].fetch(), "Some text")

    def test_assign_dict(self):
        for exp in self.get_experiments(flush_period=0.5):
            with self.subTest(msg=f"For type {exp.container_type}"):
                now = datetime.now()
                exp.assign(
                    {
                        "x": 5,
                        "metadata": {"name": "Trol", "age": 376},
                        "toys": StringSeries(["cudgel", "hat"]),
                        "nested": {"nested": {"deep_secret": FloatSeries([13, 15])}},
                        "simple_types": {
                            "int": 42,
                            "str": "imagine",
                            "float": 3.14,
                            "datetime": now,
                            "list": list(range(10)),
                        },
                    }
                )
                self.assertEqual(exp["x"].fetch(), 5)
                self.assertEqual(exp["metadata/name"].fetch(), "Trol")
                self.assertEqual(exp["metadata/age"].fetch(), 376)
                self.assertEqual(exp["toys"].fetch_last(), "hat")
                self.assertEqual(exp["nested/nested/deep_secret"].fetch_last(), 15)
                self.assertEqual(exp["simple_types/int"].fetch(), 42)
                self.assertEqual(exp["simple_types/str"].fetch(), "imagine")
                self.assertEqual(exp["simple_types/float"].fetch(), 3.14)
                self.assertEqual(
                    exp["simple_types/datetime"].fetch(),
                    now.replace(microsecond=1000 * int(now.microsecond / 1000)),
                )
                self.assertEqual(exp["simple_types/list"].fetch(), str(list(range(10))))

    def test_assign_false(self):
        # https://github.com/neptune-ai/neptune-client/issues/555
        for exp in self.get_experiments():
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp["params"] = {"predictor.cheat": False}

                self.assertFalse(exp["params/predictor.cheat"].fetch())

    def test_assign_copy(self):
        now = datetime.now()
        for exp in self.get_experiments():
            with self.subTest(msg=f"For type {exp.container_type}"):
                test_values = [
                    ("num", 42),
                    ("str", "Bat'leth"),
                    ("float", 63.2),
                    ("bool", True),
                    (
                        "datetime",
                        now.replace(microsecond=1000 * int(now.microsecond / 1000)),
                    ),
                ]
                for attr_name, attr_value in test_values:
                    exp[f"some/path/{attr_name}"] = attr_value
                    exp[f"copied/{attr_name}"] = exp[f"some/path/{attr_name}"]

                exp.wait()

                for attr_name, attr_value in test_values:
                    self.assertEqual(attr_value, exp[f"copied/{attr_name}"].fetch())

    def test_assign_copy_to_existing(self):
        for exp in self.get_experiments():
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp["some/path/num"] = 42
                exp["copied/path"] = 54
                exp["copied/path"] = exp["some/path/num"]
                exp.wait()
                self.assertEqual(42, exp["copied/path"].fetch())

    def test_assign_copy_from_nonexistent(self):
        for exp in self.get_experiments():
            with self.subTest(msg=f"For type {exp.container_type}"):
                with self.assertRaises(Exception) as e:
                    exp["copied/path"] = exp["some/path/num"]
                self.assertEqual("NoneType doesn't support copying", str(e.exception))

    def test_access_blocked_after_stop(self):
        for exp in self.get_experiments():
            with self.subTest(msg=f"For type {exp.container_type}"):
                exp["attr1"] = 1

                exp.stop()

                if isinstance(exp, Run):
                    expected_exception = InactiveRunException
                elif isinstance(exp, Project):
                    expected_exception = InactiveProjectException
                elif isinstance(exp, Model):
                    expected_exception = InactiveModelException
                elif isinstance(exp, ModelVersion):
                    expected_exception = InactiveModelVersionException
                else:
                    raise ValueError(f"Not supported exp type: {type(exp)}")

                with self.assertRaises(expected_exception):
                    exp["attr1"].fetch()
                with self.assertRaises(expected_exception):
                    exp["attr2"] = 2
                with self.assertRaises(expected_exception):
                    exp["series"].log(1)

    def test_protected_paths(self):
        model = init_model(key="MOD", mode="debug")
        model_version = init_model_version(model=model["sys/id"].fetch(), mode="debug")
        with self.assertRaises(NeptuneProtectedPathException):
            model_version["sys/stage"] = "production"

        model_version["tmp/placeholder"] = "production"
        with self.assertRaises(NeptuneProtectedPathException):
            model_version["sys/stage"] = model_version["tmp/placeholder"]

        with self.assertRaises(NeptuneProtectedPathException):
            del model_version["sys/stage"]

        with self.assertRaises(NeptuneProtectedPathException):
            model_version["sys"].pop("stage")

        with self.assertRaises(NeptuneProtectedPathException):
            del model_version["sys"]



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["get_operation_processor"]

import os
import threading
from datetime import datetime
from pathlib import Path

from neptune.new.constants import (
    ASYNC_DIRECTORY,
    NEPTUNE_DATA_DIRECTORY,
    OFFLINE_DIRECTORY,
)
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.disk_queue import DiskQueue
from neptune.new.internal.id_formats import UniqueId
from neptune.new.internal.operation import Operation
from neptune.new.sync.utils import create_dir_name
from neptune.new.types.mode import Mode

from .async_operation_processor import AsyncOperationProcessor
from .offline_operation_processor import OfflineOperationProcessor
from .operation_processor import OperationProcessor
from .read_only_operation_processor import ReadOnlyOperationProcessor
from .sync_operation_processor import SyncOperationProcessor


def get_operation_processor(
    mode: Mode,
    container_id: UniqueId,
    container_type: ContainerType,
    backend: NeptuneBackend,
    lock: threading.RLock,
    flush_period: float,
) -> OperationProcessor:
    if mode == Mode.ASYNC:
        data_path = (
            f"{NEPTUNE_DATA_DIRECTORY}/{ASYNC_DIRECTORY}"
            f"/{create_dir_name(container_type, container_id)}"
        )
        try:
            execution_id = len(os.listdir(data_path))
        except FileNotFoundError:
            execution_id = 0
        execution_path = "{}/exec-{}-{}".format(data_path, execution_id, datetime.now())
        execution_path = execution_path.replace(" ", "_").replace(":", ".")
        return AsyncOperationProcessor(
            container_id,
            container_type,
            DiskQueue(
                dir_path=Path(execution_path),
                to_dict=lambda x: x.to_dict(),
                from_dict=Operation.from_dict,
                lock=lock,
            ),
            backend,
            lock,
            sleep_time=flush_period,
        )
    elif mode == Mode.SYNC:
        return SyncOperationProcessor(container_id, container_type, backend)
    elif mode == Mode.DEBUG:
        return SyncOperationProcessor(container_id, container_type, backend)
    elif mode == Mode.OFFLINE:
        # the object was returned by mocked backend and has some random ID.
        data_path = (
            f"{NEPTUNE_DATA_DIRECTORY}/{OFFLINE_DIRECTORY}"
            f"/{create_dir_name(container_type, container_id)}"
        )
        storage_queue = DiskQueue(
            dir_path=Path(data_path),
            to_dict=lambda x: x.to_dict(),
            from_dict=Operation.from_dict,
            lock=lock,
        )
        return OfflineOperationProcessor(storage_queue)
    elif mode == Mode.READ_ONLY:
        return ReadOnlyOperationProcessor(container_id, backend)
    else:
        raise ValueError(f"mode should be one of {[m for m in Mode]}")



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import TYPE_CHECKING, Iterable, List, Optional

import click

from neptune.new.attributes.series.fetchable_series import FetchableSeries
from neptune.new.attributes.series.series import Series
from neptune.new.internal.backends.api_model import StringSeriesValues
from neptune.new.internal.operation import ClearStringLog, LogStrings, Operation
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.types.series.string_series import StringSeries as StringSeriesVal
from neptune.utils import split_to_chunks

if TYPE_CHECKING:
    from neptune.new.metadata_containers import MetadataContainer

Val = StringSeriesVal
Data = str

MAX_STRING_SERIES_VALUE_LENGTH = 1000


class StringSeries(Series[Val, Data], FetchableSeries[StringSeriesValues]):
    def __init__(self, container: "MetadataContainer", path: List[str]):
        super().__init__(container, path)
        self._value_truncation_occurred = False

    def _get_log_operations_from_value(
        self, value: Val, step: Optional[float], timestamp: float
    ) -> List[Operation]:
        values = [v[:MAX_STRING_SERIES_VALUE_LENGTH] for v in value.values]
        if not self._value_truncation_occurred and any(
            [len(v) > MAX_STRING_SERIES_VALUE_LENGTH for v in value.values]
        ):
            # the first truncation
            self._value_truncation_occurred = True
            click.echo(
                f"Warning: string series '{ path_to_str(self._path)}' value was "
                f"longer than {MAX_STRING_SERIES_VALUE_LENGTH} characters and was truncated. "
                f"This warning is printed only once per series.",
                err=True,
            )

        values = [LogStrings.ValueType(val, step=step, ts=timestamp) for val in values]
        return [LogStrings(self._path, chunk) for chunk in split_to_chunks(values, 10)]

    def _get_clear_operation(self) -> Operation:
        return ClearStringLog(self._path)

    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        if kwargs:
            click.echo(
                "Warning: unexpected arguments ({kwargs}) in StringSeries".format(
                    kwargs=kwargs
                ),
                err=True,
            )
        return StringSeriesVal(values)

    def _is_value_type(self, value) -> bool:
        return isinstance(value, StringSeriesVal)

    def fetch_last(self) -> str:
        # pylint: disable=protected-access
        val = self._backend.get_string_series_attribute(
            self._container_id, self._container_type, self._path
        )
        return val.last

    def _fetch_values_from_backend(self, offset, limit) -> StringSeriesValues:
        return self._backend.get_string_series_values(
            self._container_id, self._container_type, self._path, offset, limit
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
import unittest

from mock import patch

from neptune.new import ANONYMOUS, init_model
from neptune.new.attributes import String
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import NeptuneException, NeptuneWrongInitParametersException
from neptune.new.internal.backends.api_model import (
    Attribute,
    AttributeType,
    IntAttribute,
)
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from tests.neptune.new.client.abstract_experiment_test_mixin import (
    AbstractExperimentTestMixin,
)
from tests.neptune.new.utils.api_experiments_factory import api_model

AN_API_MODEL = api_model()


@patch("neptune.new.internal.backends.factory.HostedNeptuneBackend", NeptuneBackendMock)
class TestClientModel(AbstractExperimentTestMixin, unittest.TestCase):
    @staticmethod
    def call_init(**kwargs):
        return init_model(key="MOD", **kwargs)

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    def test_offline_mode(self):
        with self.assertRaises(NeptuneException):
            init_model(key="MOD", mode="offline")

    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_metadata_container",
        new=lambda _, container_id, expected_container_type: AN_API_MODEL,
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
        new=lambda _, _uuid, _type: [Attribute("some/variable", AttributeType.INT)],
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_int_attribute",
        new=lambda _, _uuid, _type, _path: IntAttribute(42),
    )
    def test_read_only_mode(self):
        exp = init_model(mode="read-only", model="whatever")

        with self.assertLogs() as caplog:
            exp["some/variable"] = 13
            exp["some/other_variable"] = 11
            self.assertEqual(
                caplog.output,
                [
                    "WARNING:neptune.new.internal.operation_processors.read_only_operation_processor:"
                    "Client in read-only mode, nothing will be saved to server."
                ],
            )

        self.assertEqual(42, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_metadata_container",
        new=lambda _, container_id, expected_container_type: AN_API_MODEL,
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
        new=lambda _, _uuid, _type: [Attribute("test", AttributeType.STRING)],
    )
    def test_resume(self):
        with init_model(flush_period=0.5, model="whatever") as exp:
            self.assertEqual(exp._id, AN_API_MODEL.id)
            self.assertIsInstance(exp.get_structure()["test"], String)

    def test_wrong_parameters(self):
        with self.assertRaises(NeptuneWrongInitParametersException):
            init_model(model=None, key=None)

    def test_name_parameter(self):
        with init_model(key="TRY", name="some_name") as exp:
            exp.wait()
            self.assertEqual(exp["sys/name"].fetch(), "some_name")



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import unittest
from typing import List

import mock
from freezegun import freeze_time
from mock import MagicMock

from neptune.internal.api_clients import HostedNeptuneBackendApiClient
from neptune.internal.api_clients.hosted_api_clients.hosted_alpha_leaderboard_api_client import (
    HostedAlphaLeaderboardApiClient,
)
from neptune.internal.channels.channels import (
    ChannelIdWithValues,
    ChannelNamespace,
    ChannelType,
    ChannelValue,
)
from tests.neptune.new.backend_test_mixin import (
    BackendTestMixin as AlphaBackendTestMixin,
)

API_TOKEN = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYWxwaGEuc3RhZ2UubmVwdHVuZS5haSIsImFwaV91cmwiOiJodHRwczovL2FscG"
    "hhLnN0YWdlLm5lcHR1bmUuYWkiLCJhcGlfa2V5IjoiZDg5MGQ3Y2ItZGEzNi00MjRkLWJhNTQtZmVjZDJmYTdhOTQzIn0="
)
"""base64 decoded `API_TOKEN`
{
  "api_address": "https://alpha.stage.neptune.ai",
  "api_url": "https://alpha.stage.neptune.ai",
  "api_key": "d890d7cb-da36-424d-ba54-fecd2fa7a943"
}
"""


class TestAlphaIntegrationNeptuneBackend(unittest.TestCase, AlphaBackendTestMixin):
    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch(
        "neptune.internal.api_clients.hosted_api_clients.hosted_backend_api_client.NeptuneAuthenticator",
        new=MagicMock,
    )
    @mock.patch(
        "neptune.new.internal.backends.hosted_client.NeptuneAuthenticator",
        new=MagicMock,
    )
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def setUp(self, swagger_client_factory) -> None:
        # pylint:disable=arguments-differ
        self._get_swagger_client_mock(swagger_client_factory)
        self.backend = HostedNeptuneBackendApiClient(API_TOKEN)
        self.leaderboard = HostedAlphaLeaderboardApiClient(self.backend)
        self.exp_mock = MagicMock(internal_id="00000000-0000-0000-0000-000000000000")

    @freeze_time()
    def _test_send_channel_values(
        self,
        channel_y_elements: List[tuple],
        expected_operation: str,
        channel_type: ChannelType,
    ):
        # given prepared `ChannelIdWithValues`
        channel_id = "channel_id"
        channel_name = "channel_name"
        now_ms = int(time.time() * 1000)
        channel_with_values = ChannelIdWithValues(
            channel_id=channel_id,
            channel_name=channel_name,
            channel_type=channel_type.value,
            channel_namespace=ChannelNamespace.USER,
            channel_values=[
                ChannelValue(x=None, y={channel_y_key: channel_y_value}, ts=None)
                for channel_y_key, channel_y_value in channel_y_elements
            ],
        )

        # invoke send_channels_values
        self.leaderboard.send_channels_values(self.exp_mock, [channel_with_values])

        # expect `executeOperations` was called once with properly prepared kwargs
        expected_call_args = {
            "experimentId": "00000000-0000-0000-0000-000000000000",
            "operations": [
                {
                    "path": f"logs/{channel_name}",
                    expected_operation: {
                        "entries": [
                            {
                                "value": channel_y_value,
                                "step": None,
                                "timestampMilliseconds": now_ms,
                            }
                            for _, channel_y_value in channel_y_elements
                        ]
                    },
                }
            ],
        }
        # pylint:disable=protected-access
        execute_operations = (
            self.leaderboard.leaderboard_swagger_client.api.executeOperations
        )
        self.assertEqual(len(execute_operations.call_args_list), 1)
        self.assertDictEqual(
            execute_operations.call_args_list[0][1], expected_call_args
        )

    def test_send_channels_text_values(self):
        channel_y_elements = [
            ("text_value", "Line of text"),
            ("text_value", "Another line of text"),
        ]
        self._test_send_channel_values(
            channel_y_elements,
            expected_operation="logStrings",
            channel_type=ChannelType.TEXT,
        )

    def test_send_channels_numeric_values(self):
        channel_y_elements = [
            ("numeric_value", 42),
            ("numeric_value", 0.07),
        ]
        self._test_send_channel_values(
            channel_y_elements,
            expected_operation="logFloats",
            channel_type=ChannelType.NUMERIC,
        )

    def test_send_channels_image_values(self):
        """TODO: implement in NPT-9207"""



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
import uuid
from abc import abstractmethod
from datetime import datetime

from mock import Mock, patch

from neptune.new import ANONYMOUS
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.backends.api_model import (
    Attribute,
    AttributeType,
    AttributeWithProperties,
    LeaderboardEntry,
)
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock


@patch(
    "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
    new=lambda _, _uuid, _type: [Attribute(path="test", type=AttributeType.STRING)],
)
@patch("neptune.new.internal.backends.factory.HostedNeptuneBackend", NeptuneBackendMock)
class AbstractTablesTestMixin:
    expected_container_type = None

    @abstractmethod
    def get_table(self):
        pass

    @abstractmethod
    def get_table_entries(self, table):
        pass

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    @classmethod
    def setUp(cls) -> None:
        if PROJECT_ENV_NAME in os.environ:
            del os.environ[PROJECT_ENV_NAME]

    @staticmethod
    def build_attributes_leaderboard(now: datetime):
        attributes = []
        attributes.append(
            AttributeWithProperties(
                "run/state", AttributeType.RUN_STATE, Mock(value="idle")
            )
        )
        attributes.append(
            AttributeWithProperties("float", AttributeType.FLOAT, Mock(value=12.5))
        )
        attributes.append(
            AttributeWithProperties(
                "string", AttributeType.STRING, Mock(value="some text")
            )
        )
        attributes.append(
            AttributeWithProperties("datetime", AttributeType.DATETIME, Mock(value=now))
        )
        attributes.append(
            AttributeWithProperties(
                "float/series", AttributeType.FLOAT_SERIES, Mock(last=8.7)
            )
        )
        attributes.append(
            AttributeWithProperties(
                "string/series", AttributeType.STRING_SERIES, Mock(last="last text")
            )
        )
        attributes.append(
            AttributeWithProperties(
                "string/set", AttributeType.STRING_SET, Mock(values=["a", "b"])
            )
        )
        attributes.append(
            AttributeWithProperties(
                "git/ref",
                AttributeType.GIT_REF,
                Mock(commit=Mock(commitId="abcdef0123456789")),
            )
        )
        attributes.append(AttributeWithProperties("file", AttributeType.FILE, None))
        attributes.append(
            AttributeWithProperties("file/set", AttributeType.FILE_SET, None)
        )
        attributes.append(
            AttributeWithProperties("image/series", AttributeType.IMAGE_SERIES, None)
        )
        return attributes

    @patch.object(NeptuneBackendMock, "search_leaderboard_entries")
    def test_get_table_as_pandas(self, search_leaderboard_entries):
        # given
        now = datetime.now()
        attributes = self.build_attributes_leaderboard(now)

        # and
        empty_entry = LeaderboardEntry(str(uuid.uuid4()), [])
        filled_entry = LeaderboardEntry(str(uuid.uuid4()), attributes)
        search_leaderboard_entries.return_value = [empty_entry, filled_entry]

        # when
        df = self.get_table().to_pandas()

        # then
        self.assertEqual("idle", df["run/state"][1])
        self.assertEqual(12.5, df["float"][1])
        self.assertEqual("some text", df["string"][1])
        self.assertEqual(now, df["datetime"][1])
        self.assertEqual(8.7, df["float/series"][1])
        self.assertEqual("last text", df["string/series"][1])
        self.assertEqual("a,b", df["string/set"][1])
        self.assertEqual("abcdef0123456789", df["git/ref"][1])

        with self.assertRaises(KeyError):
            self.assertTrue(df["file"])
        with self.assertRaises(KeyError):
            self.assertTrue(df["file/set"])
        with self.assertRaises(KeyError):
            self.assertTrue(df["image/series"])

    @patch.object(NeptuneBackendMock, "search_leaderboard_entries")
    @patch.object(NeptuneBackendMock, "download_file")
    @patch.object(NeptuneBackendMock, "download_file_set")
    def test_get_table_as_table_entries(
        self,
        download_file_set,
        download_file,
        search_leaderboard_entries,
    ):
        # given
        exp_id = str(uuid.uuid4())
        now = datetime.now()
        attributes = self.build_attributes_leaderboard(now)

        # and
        search_leaderboard_entries.return_value = [LeaderboardEntry(exp_id, attributes)]

        # when
        table_entry = self.get_table_entries(table=self.get_table())[0]

        # then
        self.assertEqual("idle", table_entry["run/state"].get())
        self.assertEqual("idle", table_entry["run"]["state"].get())
        self.assertEqual(12.5, table_entry["float"].get())
        self.assertEqual("some text", table_entry["string"].get())
        self.assertEqual(now, table_entry["datetime"].get())
        self.assertEqual(8.7, table_entry["float/series"].get())
        self.assertEqual("last text", table_entry["string/series"].get())
        self.assertEqual({"a", "b"}, table_entry["string/set"].get())
        self.assertEqual("abcdef0123456789", table_entry["git/ref"].get())

        with self.assertRaises(MetadataInconsistency):
            table_entry["file"].get()
        with self.assertRaises(MetadataInconsistency):
            table_entry["file/set"].get()
        with self.assertRaises(MetadataInconsistency):
            table_entry["image/series"].get()

        table_entry["file"].download("some_directory")
        download_file.assert_called_with(
            container_id=exp_id,
            container_type=self.expected_container_type,
            path=["file"],
            destination="some_directory",
        )

        table_entry["file/set"].download("some_directory")
        download_file_set.assert_called_with(
            container_id=exp_id,
            container_type=self.expected_container_type,
            path=["file", "set"],
            destination="some_directory",
        )
